from nagra.transaction import ExecMany


def test_resolver(transaction, person):
    placeholder = "?" if transaction.flavor == "sqlite" else "%s"
    values = [(i,) for i in range(5)]
    rsv = ExecMany(
        f"SELECT {placeholder}",
        values,
        trn=transaction,
    )
    results = list(rsv)
    assert results == [(0,), (1,), (2,), (3,), (4,)]

    rsv = ExecMany(
        f"SELECT {placeholder} from person WHERE false",
        values,
        trn=transaction,
    )
    results = list(rsv)
    assert results == [None] * 5



from typing import Union, Optional, TYPE_CHECKING
from collections.abc import Iterable
from dataclasses import dataclass

try:
    from pandas import DataFrame
except ImportError:
    DataFrame = None

from nagra import Statement, Schema
from nagra.transaction import Transaction
from nagra.writer import WriterMixin
from nagra.utils import snake_to_pascal, get_table_from_dataclass, iter_dataclass_cols

if TYPE_CHECKING:
    from nagra.table import Table, Env


class Upsert(WriterMixin):
    def __init__(
        self,
        table: "Table",
        *columns: str,
        trn: Transaction,
        env: "Env",
        lenient: Union[bool, list[str], None] = None,
        insert_only: bool = False,
        check: Iterable[str] = [],
    ):
        self.table = table
        self.columns = [c.lstrip(".") for c in columns]
        self._insert_only = insert_only
        self.lenient = lenient or []
        self._check = list(check)
        self.trn = trn
        self.env = env
        super().__init__()

    def clone(
        self,
        trn: Optional["Transaction"] = None,
        insert_only: Optional[bool] = None,
        check: Iterable[str] = [],
    ):
        """
        Return a copy of upsert with updated parameters
        """
        trn = trn or self.trn
        insert_only = self._insert_only if insert_only is None else insert_only
        check = self._check + list(check)
        cln = Upsert(
            self.table,
            *self.columns,
            trn=trn,
            env=self.env.clone(),
            lenient=self.lenient,
            insert_only=insert_only,
            check=check,
        )
        return cln

    def insert_only(self):
        return self.clone(insert_only=True)

    def check(self, *conditions: str):
        return self.clone(check=conditions)

    def stm(self):
        pk = self.table.primary_key
        conflict_key = [pk] if pk in self.groups else self.table.natural_key
        columns = self.groups
        do_update = False if self._insert_only else len(columns) > len(conflict_key)
        stm = Statement(
            "upsert",
            self.trn.flavor,
            table=self.table.name,
            columns=columns,
            conflict_key=conflict_key,
            do_update=do_update,
            returning=[pk] if pk else self.table.natural_key,
        )
        return stm()

    def _exec_args(self, arg_df):
        args = zip(*(arg_df[c] for c in self.groups))
        return args

    def resolve(self, column: str, *values: list[any]):
        rows = list(zip(*values))
        yield from self._resolve(column, rows)

    @staticmethod
    def from_dataclass(cls: dataclass, schema: Schema = Schema.default) -> "Upsert":
        # TODO add from_pydantic
        table = get_table_from_dataclass(cls, schema)
        cols = list(iter_dataclass_cols(cls))
        return table.upsert(*cols)



from importlib.metadata import version

from .statement import Statement
from .schema import Schema
from .table import Table
from .view import View
from .transaction import Transaction


__version__ = version("nagra")



import datetime
import pytest

from nagra import Transaction
from nagra.utils import strip_lines


def test_simple_select(person):
    stm = person.select("name").stm()
    res = " ".join(strip_lines(stm))
    assert res == 'SELECT "person"."name" FROM "person" ;'


def test_select_with_join(person):
    stm = person.select("name", "parent.parent.name").stm()
    res = strip_lines(stm)
    assert res == [
        "SELECT",
        '"person"."name", "parent_1"."name"',
        'FROM "person"',
        'LEFT JOIN "person" as parent_0 ON (',
        'parent_0."id" = "person"."parent"',
        ")",
        'LEFT JOIN "person" as parent_1 ON (',
        'parent_1."id" = "parent_0"."parent"',
        ")",
        ";",
    ]


def test_kitchensink_select(kitchensink):
    stm = kitchensink.select().stm()
    res = strip_lines(stm)
    assert res == [
        "SELECT",
        '"kitchensink"."varchar", "kitchensink"."bigint", "kitchensink"."float", '
        '"kitchensink"."int", "kitchensink"."timestamp", "kitchensink"."timestamptz", '
        '"kitchensink"."bool", "kitchensink"."date", "kitchensink"."json", '
        '"kitchensink"."uuid", "kitchensink"."max", "kitchensink"."true", '
        '"kitchensink"."blob"',
        'FROM "kitchensink"',
        ";",
    ]

    stm = kitchensink.select(".true").stm()
    res = strip_lines(stm)
    assert res == ["SELECT", '"kitchensink"."true"', 'FROM "kitchensink"', ";"]

    stm = kitchensink.select("true").stm()
    res = strip_lines(stm)
    assert res == ["SELECT", "true", 'FROM "kitchensink"', ";"]


def test_select_clone(person):
    queries = [
        person.select("name")
        .limit(1)
        .offset(1)
        .where("(= name 'spam')")
        .groupby("name"),
        person.select("name")
        .offset(1)
        .where("(= name 'spam')")
        .groupby("name")
        .limit(1),
        person.select("name")
        .where("(= name 'spam')")
        .groupby("name")
        .limit(1)
        .offset(1),
        person.select("name")
        .groupby("name")
        .limit(1)
        .offset(1)
        .where("(= name 'spam')"),
    ]

    expected = queries[0].stm()
    for q in queries[1:]:
        assert q.stm() == expected


def test_select_with_where(person):
    select = person.select("name").where("(= id {})")
    stm = select.stm()
    res = strip_lines(stm)
    assert res == [
        "SELECT",
        '"person"."name"',
        'FROM "person"',
        "WHERE",
        '"person"."id" = %s',
        ";",
    ]

    stm = select.where("(= name 'spam')").stm()
    res = strip_lines(stm)
    assert res == [
        "SELECT",
        '"person"."name"',
        'FROM "person"',
        "WHERE",
        '"person"."id" = %s AND "person"."name" = \'spam\'',
        ";",
    ]


def test_select_where_and_join(person):
    select = person.select("name").where("(= parent.name 'foo')")
    stm = select.stm()
    res = strip_lines(stm)
    assert res == [
        "SELECT",
        '"person"."name"',
        'FROM "person"',
        'LEFT JOIN "person" as parent_0 ON (',
        'parent_0."id" = "person"."parent"',
        ")",
        "WHERE",
        '"parent_0"."name" = \'foo\'',
        ";",
    ]


@pytest.mark.parametrize("op", ["min", "max", "sum"])
def test_simple_agg(person, op):
    # MIN
    stm = person.select(f"({op} name)").stm()
    res = " ".join(strip_lines(stm))
    assert res == f'SELECT {op}("person"."name") FROM "person" ;'


def test_count(person):
    stm = person.select("(count *)").stm()
    res = " ".join(strip_lines(stm))
    assert res == 'SELECT count(*) FROM "person" ;'

    stm = person.select("(count 1)").stm()
    res = " ".join(strip_lines(stm))
    assert res == 'SELECT count(1) FROM "person" ;'


def test_groupby(person):
    # Explicit
    stm = person.select("name", "(count *)").groupby("name").stm()
    res = " ".join(strip_lines(stm))
    assert (
        res
        == 'SELECT "person"."name", count(*) FROM "person" GROUP BY "person"."name" ;'
    )

    # implicit
    stm = person.select("name", "(count *)").stm()
    res = " ".join(strip_lines(stm))
    assert (
        res
        == 'SELECT "person"."name", count(*) FROM "person" GROUP BY "person"."name" ;'
    )


def test_orderby(person):
    # asc
    stm = person.select("name").orderby("name").stm()
    res = " ".join(strip_lines(stm))
    assert res == 'SELECT "person"."name" FROM "person" ORDER BY "person"."name" asc ;'

    # desc
    stm = person.select("name").orderby(("name", "desc")).stm()
    res = " ".join(strip_lines(stm))
    assert res == 'SELECT "person"."name" FROM "person" ORDER BY "person"."name" desc ;'

    # with join
    stm = person.select("name").orderby("parent.name").stm()
    res = " ".join(strip_lines(stm))
    assert res == (
        'SELECT "person"."name" FROM "person" LEFT JOIN "person" as parent_0 ON ( '
        'parent_0."id" = "person"."parent" ) ORDER BY "parent_0"."name" asc ;'
    )


def test_o2m_stm(person, org):
    # Combine one2many and implicit joins
    select = person.select(
        "name",
        "orgs.name",
        "parent.name",
        "parent.parent.name",
    )
    stm = select.stm()
    res = strip_lines(stm)
    expected = [
        "SELECT",
        '"person"."name", "orgs_0"."name", "parent_1"."name", ' '"parent_2"."name"',
        'FROM "person"',
        'LEFT JOIN "org" as orgs_0 ON (',
        'orgs_0."person" = "person"."id"',
        ")",
        'LEFT JOIN "person" as parent_1 ON (',
        'parent_1."id" = "person"."parent"',
        ")",
        'LEFT JOIN "person" as parent_2 ON (',
        'parent_2."id" = "parent_1"."parent"',
        ")",
        ";",
    ]
    assert res == expected

    # Multiple one2many
    select = person.select(
        "name",
        "orgs.country",
        "skills.name",
    )
    stm = select.stm()
    res = strip_lines(stm)
    expected = [
        "SELECT",
        '"person"."name", "orgs_0"."country", "skills_1"."name"',
        'FROM "person"',
        'LEFT JOIN "org" as orgs_0 ON (',
        'orgs_0."person" = "person"."id"',
        ")",
        'LEFT JOIN "skill" as skills_1 ON (',
        'skills_1."person" = "person"."id"',
        ")",
        ";",
    ]
    assert res == expected

    # Use a on2many after a many2one
    select = org.select(
        "name",
        "person.name",
        "person.skills.name",
    )
    stm = select.stm()
    res = strip_lines(stm)
    expected = [
        "SELECT",
        '"org"."name", "person_0"."name", "skills_1"."name"',
        'FROM "org"',
        'LEFT JOIN "person" as person_0 ON (',
        'person_0."id" = "org"."person"',
        ")",
        'LEFT JOIN "skill" as skills_1 ON (',
        'skills_1."person" = "person_0"."id"',
        ")",
        ";",
    ]
    assert res == expected

    # Use a on2many after a one2many
    select = person.select("name", "orgs.addresses.city")
    stm = select.stm()
    res = strip_lines(stm)
    expected = [
        "SELECT",
        '"person"."name", "addresses_1"."city"',
        'FROM "person"',
        'LEFT JOIN "org" as orgs_0 ON (',
        'orgs_0."person" = "person"."id"',
        ")",
        'LEFT JOIN "address" as addresses_1 ON (',
        'addresses_1."org" = "orgs_0"."id"',
        ")",
        ";",
    ]
    assert res == expected


def test_o2m_select(transaction, person, org, address):
    # Test with actual data
    person.upsert("name").execute("Charly")
    org.upsert("name", "person.name").execute("Alpha", "Charly")
    org.upsert("name", "person.name").execute("Beta", "Charly")
    address.upsert("city", "org.name").executemany(
        [
            ("Ankara", "Alpha"),
            ("Athens", "Alpha"),
            ("Beirut", "Beta"),
        ]
    )
    rows = list(
        person.select("name", "orgs.addresses.city").orderby("orgs.addresses.city")
    )
    assert rows == [("Charly", "Ankara"), ("Charly", "Athens"), ("Charly", "Beirut")]


def test_agg(transaction, temperature):
    temperature.upsert("timestamp", "city", "value").executemany(
        [
            ("1970-01-01", "Berlin", 10),
            ("1970-01-01", "London", 12),
        ]
    )
    rows = list(
        temperature.select(
            "city",
        )
    )
    assert len(rows) == 2

    # String concat
    is_pg = Transaction.current().flavor == "postgresql"
    if is_pg:
        select = temperature.select("(string_agg city ',')")
    else:
        select = temperature.select("(group_concat city)")
    rows = list(select)
    assert len(rows) == 1
    (record,) = rows
    assert record[0] in ("Berlin,London", "London,Berlin")

    # Strings into array
    if is_pg:
        (record,) = list(temperature.select("(array_agg city)"))
        assert sorted(record[0]) == ["Berlin", "London"]

    # sum, avg, min and max
    for op, expected in [("sum", 22), ("min", 10), ("max", 12), ("avg", 11)]:
        select = temperature.select(f"({op} value)")
        (record,) = list(select)
        assert expected == record[0]

    # Add more rows
    temperature.upsert("timestamp", "city", "value").executemany(
        [
            ("1970-01-02", "Berlin", 10),
            ("1970-01-02", "London", 12),
        ]
    )

    records = dict(temperature.select("city", "(sum value)").groupby("city"))
    assert records == {"Berlin": 20.0, "London": 24.0}

    # Json agg
    if is_pg:
        select = temperature.select("(json_object_agg city value)")
        (record,) = list(select)
        assert record[0] == {"Berlin": 10, "London": 12}


def test_date_op(transaction, temperature):
    is_pg = Transaction.current().flavor == "postgresql"

    temperature.upsert("timestamp", "city", "value").executemany(
        [
            ("1970-01-02", "Berlin", 10),
            ("1970-01-02", "London", 12),
        ]
    )
    if is_pg:
        select = temperature.select("(extract 'year' timestamp)")
        records = list(select)
        assert records[0][0] == 1970
    else:
        select = temperature.select("(strftime '%Y' timestamp)")
        records = list(select)
        assert records[0][0] == "1970"
    assert len(records) == 2


def test_to_dict(transaction, temperature):
    # Upsert
    temperature.upsert("timestamp", "city", "value").executemany(
        [
            ("1970-01-02", "Berlin", 10),
            ("1970-01-02", "London", 12),
        ]
    )
    # Read data
    expected_date = datetime.datetime(1970, 1, 2, 0, 0)
    if transaction.flavor == "sqlite":
        expected_date = str(expected_date.date())

    records = list(temperature.select().orderby("city").to_dict())
    assert len(records) == 2
    assert records[0] == {
        "timestamp": expected_date,
        "city": "Berlin",
        "value": 10.0,
    }
    # Read with custom arg
    cond = "(= value {})"
    (record,) = temperature.select().where(cond).to_dict(12)
    assert record == {
        "timestamp": expected_date,
        "city": "London",
        "value": 12.0,
    }


def test_select_alias(transaction, temperature):
    # Upsert
    temperature.upsert("timestamp", "city", "value").executemany(
        [
            ("1970-01-02", "Berlin", 10),
            ("1970-01-02", "London", 12),
        ]
    )
    select = (
        temperature.select()
        .orderby("city")
        .aliases("t", "c", "v")
    )

    # Check record keys
    records = list(select.to_dict())
    assert sorted(records[0]) == ["c", "t", "v"]

    # Check pandas df columns
    df = select.to_pandas()
    assert sorted(df) == ["c", "t", "v"]

    # Check polars df columns
    if transaction.flavor == "postgresql":
        df = select.to_polars().collect()
        assert sorted(df.columns) == ["c", "t", "v"]


@pytest.mark.parametrize("nest_with_param", [False, True])
def test_to_nested_dict(transaction, person, nest_with_param):
    # Upsert
    person.upsert("name", "parent.name").execute("Alice", None)
    person.upsert("name", "parent.name").execute("Bob", "Alice")
    person.upsert("name", "parent.name").execute("Charly", "Bob")

    # Read data
    select = person.select("name", "parent.name", "parent.parent.name")
    if nest_with_param:
        records = list(select.to_dict(nest=True))
    else:
        records = list(select.to_nested_dict())

    assert records == [
        {
            "name": "Alice",
            "parent": None,
        },
        {
            "name": "Bob",
            "parent": {
                "name": "Alice",
                "parent": None,
            },
        },
        {
            "name": "Charly",
            "parent": {
                "name": "Bob",
                "parent": {
                    "name": "Alice",
                },
            },
        },
    ]



from itertools import islice
from typing import Iterable, Union, TYPE_CHECKING

from nagra.transaction import Transaction


if TYPE_CHECKING:
    from nagra.table import Table


def copy_from(
    table: "Table",
    rows: Iterable[tuple],
    trn: Transaction,
    lenient: Union[bool, list[str], None] = None,
):
    """
    Populate table with a COPY FROM statement.
    """

    if trn.flavor == "sqlite":
        raise NotImplementedError("COPY FROM not available for sqlite")

    stm = f'COPY "{table.name}" FROM STDIN'
    cursor = trn.connection.cursor()
    it = iter(rows)
    with cursor.copy(stm) as copy:
        while True:
            chunk = list(islice(it, 1000))
            if not chunk:
                break
            content = serialize_chunk(chunk, table.columns)
            copy.write(content)


def serialize_chunk(rows, columns):
    row_strs = ("\t".join(stringify(v) for v in row) for row in rows)
    return "\n".join(row_strs) + "\n"


def stringify(value):
    """
    Convert value to string for COPY FROM
    """
    if isinstance(value, str):
        return f'"{value}"'
    elif isinstance(value, bool):
        return "true" if value else "false"
    elif value is None:
        return "\\N"
    else:
        return str(value)



import zoneinfo
from datetime import datetime, date
from uuid import UUID

import pytest
from pandas import DataFrame, to_datetime

from nagra.exceptions import UnresolvedFK, ValidationError


def test_simple_update_by_id(transaction, person):
    # First upsert some values
    upsert = person.upsert("name")
    new_id = upsert.execute("Bob")

    # Update by id
    update = person.update("id", "name")
    new_id_copy = update.execute(new_id, "BOB")
    assert new_id_copy == new_id

    # Test update is successful
    (row,) = person.select("id", "name")
    assert row == (new_id, "BOB")


def test_simple_update_by_nk(transaction, temperature):
    # First upsert some values
    upsert = temperature.upsert("timestamp", "city", "value")
    new_id = upsert.execute("2024-06-27 17:52:00", "Brussels", 27)

    # Update by nk
    update = temperature.update("timestamp", "city", "value")
    new_id_copy = update.execute("2024-06-27 17:52:00", "Brussels", 28)
    assert new_id_copy == new_id

    # Test update is successful
    (row,) = temperature.select("timestamp", "city", "value")
    ts = "2024-06-27 17:52:00"
    if transaction.flavor != "sqlite":
        ts = datetime.fromisoformat(ts)
    assert row == (ts, "Brussels", 28)


def test_return_ids(transaction, person):
    # Add lines
    upsert = person.upsert("name", "parent.name")
    records = [("Alice", None), ("Bob", None), ("Charlie", None)]
    insert_ids = upsert.executemany(records)

    # Fisrt updates - by nk
    update = person.update("name", "parent")
    pid = insert_ids[0]
    records = [("Bob", pid), ("Charlie", pid)]
    update_ids = update.executemany(records)
    assert update_ids == insert_ids[1:]

    # Second update - by id
    update = person.update("name", "id")
    a, b, c = insert_ids
    records = [("BOB", b), ("CHARLIE", c)]
    update_ids = update.executemany(records)
    assert update_ids == insert_ids[1:]

    # Check consistency
    rows = list(person.select("name").where(f"(= parent {a}").orderby("id"))
    assert rows == [("BOB",), ("CHARLIE",)]


def test_from_pandas(transaction, kitchensink):
    # First insert
    df = DataFrame(
        {
            "varchar": ["ham"],
            "bigint": [1],
            "float": [1.0],
            "int": [1],
            "timestamp": to_datetime(["1970-01-01 00:00:00"]),
            "timestamptz": to_datetime(["1970-01-01 00:00:00+00:00"]),
            "bool": [True],
            "date": ["1970-01-01"],
            "json": [{}],
            "uuid": ["F1172BD3-0A1D-422E-8ED6-8DC2D0F8C11C"],
            "max": ["max"],
            "true": ["true"],
            "blob": [b"blob"],
        }
    )
    ids = kitchensink.upsert().from_pandas(df)

    # Update values
    df = DataFrame(
        {
            "id": ids,
            "varchar": ["HAM"],
            "bigint": [2],
            "float": [2.0],
            "int": [2],
            "timestamp": to_datetime(["1970-01-02 00:00:00"]),
            "timestamptz": to_datetime(["1970-01-02 00:00:00+00:00"]),
            "bool": [False],
            "date": ["1970-01-02"],
            "json": '[{"foo": "bar"}]',
            "uuid": ["F1172BD3-0A1D-422E-8ED6-8DC2D0F8C11D"],
            "max": ["max"],
            "true": ["true"],
        }
    )
    kitchensink.update(*df.columns).from_pandas(df)

    (row,) = kitchensink.select()
    BRUTZ = zoneinfo.ZoneInfo(key="Europe/Brussels")
    if transaction.flavor == "postgresql":
        assert row == (
            "HAM",
            2,
            2.0,
            2,
            datetime(1970, 1, 2, 0, 0),
            datetime(1970, 1, 2, 1, 0, tzinfo=BRUTZ),
            False,
            date(1970, 1, 2),
            [{"foo": "bar"}],
            UUID("F1172BD3-0A1D-422E-8ED6-8DC2D0F8C11D"),
            "max",
            "true",
            b"blob",
        )
    else:
        assert row == (
            "HAM",
            2,
            2.0,
            2,
            "1970-01-02",
            "1970-01-02 00:00:00+00:00",
            0,
            "1970-01-02",
            '[{"foo": "bar"}]',
            "F1172BD3-0A1D-422E-8ED6-8DC2D0F8C11D",
            "max",
            "true",
            "blob",
        )


def test_where_cond(transaction, person):
    """
    Shows that an exception is raised when a row infrige a check condition
    """
    upsert = person.upsert("name")
    upsert.execute("Tango")

    cond = "(!= name parent.name)"  # Forbid self-reference
    upsert = person.update("name", "parent.name").check(cond)
    with pytest.raises(ValidationError):
        upsert.execute("Tango", "Tango")


def test_missing_fk(transaction, person):
    # If pass None in parent.name, we get None back
    upsert = person.upsert("name", "parent.name")
    records = [("Big Alice", None), ("Big Bob", None)]
    upsert.executemany(records)

    # If given a non-existing name update raises UnresolvedFK exception
    update = person.update("name", "parent.name")
    records = [("Big Alice", "I do not exist")]
    with pytest.raises(UnresolvedFK):
        update.executemany(records)

    # If lenient is given a None is inserted
    for lenient in [True, ["parent"]]:
        update = person.update("name", "parent.name", lenient=lenient)
        records = [("Big Alice", "I do not exist")]
        update.executemany(records)
        rows = list(person.select("parent").where("(= name 'Big Alice')").execute())
        assert rows == [(None,)]



import csv
import os
import bisect
import logging
import re
import sys
from contextlib import contextmanager
from dataclasses import dataclass, fields, is_dataclass
from pathlib import Path
from time import perf_counter
from typing import Iterator, TYPE_CHECKING, get_args
from enum import StrEnum

from jinja2 import FileSystemLoader, Environment
from rich import box
from rich.console import Console
from rich.markup import escape
import rich.table

if TYPE_CHECKING:
    from nagra.schema import Schema

HERE = Path(__file__).parent
fmt = "%(levelname)s:%(asctime).19s: %(message)s"
logging.basicConfig(format=fmt)
logger = logging.getLogger("nagra")
if os.environ.get("NAGRA_DEBUG"):
    logger.setLevel("DEBUG")
    logger.debug("Log level set to debug")
UNSET = object()
RE_SC_PC = re.compile(r"(?:^|_)(\w)")
RE_PC_SC = re.compile(r"(?<!^)(?=[A-Z])")


def autoquote(x):
    if x.startswith('"'):
        return x
    return f'"{x}"'


# Setup jinja env
jinja_env = Environment(loader=FileSystemLoader(HERE / "template"))
jinja_env.filters["autoquote"] = autoquote


def snake_to_pascal(name: str) -> str:
    return RE_SC_PC.sub(lambda m: m.group(1).upper(), name)


def pascal_to_snake(name: str) -> str:
    # PascalCase to snake_case
    return RE_PC_SC.sub("_", name).lower()


def template(name):
    return jinja_env.get_template(name)


def strip_lines(stmt):
    res = []
    for r in stmt.splitlines():
        line = r.strip()
        if line:
            res.append(line)
    return res


def pretty_nb(number):
    prefixes = "yzafpnum_kMGTPEZY"
    factors = [1000**i for i in range(-8, 8)]
    if number == 0:
        return 0
    if number < 0:
        return "-" + pretty_nb(-number)
    idx = bisect.bisect_right(factors, number) - 1
    prefix = prefixes[idx]
    return "%.2f%s" % (number / factors[idx], "" if prefix == "_" else prefix)


@contextmanager
def timeit(title=""):
    start = perf_counter()
    yield
    delta = perf_counter() - start
    print(title, pretty_nb(delta) + "s", file=sys.stderr)


class TableFmt(StrEnum):
    RICH = "rich"
    CSV = "csv"


def print_table(rows, headers, pivot=False, format:TableFmt = None):
    if format == TableFmt.CSV:
        writer = csv.writer(sys.stdout)
        writer.writerow(headers)
        for row in rows:
            writer.writerow(row)
        return

    console = Console()
    escstr = lambda s: escape(str(s))
    if pivot:
        for row in rows:
            table = rich.table.Table(box=box.SIMPLE)
            for pivot_row in zip(headers, row):
                table.add_row(*map(escstr, pivot_row))
            console.print(table)
        return

    table = rich.table.Table(*headers, box=box.ROUNDED)
    for row in rows:
        table.add_row(*map(escstr, row))
    console.print(table)


def iter_dataclass_cols(cls: dataclass, prefix: str = "") -> Iterator[str]:
    """
    Return columns corresponding to the dataclass, including
    nested dataclasses (using dotted notation).
    """
    for field in fields(cls):
        name = field.name
        type = field.type
        column = f"{prefix}.{name}" if prefix else name
        sub_types = get_args(type)
        # Handle optional dataclasses
        for sub_type in sub_types:
            if is_dataclass(sub_type):
                yield from iter_dataclass_cols(sub_type, prefix=column)
                break
        else:
            # Handle other types
            if is_dataclass(type):
                yield from iter_dataclass_cols(type, prefix=column)
            else:
                yield column


def get_table_from_dataclass(cls: dataclass, schema: "Schema"):
    assert is_dataclass(cls)
    cls_name = getattr(cls, "__table__", cls.__name__)
    table = schema.get(pascal_to_snake(cls_name))
    return table



from nagra import Statement


def test_debug_statement():
    stmt = Statement("debug")
    doc = stmt()
    assert doc == ""

    stmt = stmt.foo("bar")
    doc = stmt()
    assert doc == "foo=bar\n"

    stmt = stmt.ham("spam")
    doc = stmt()
    assert doc == "foo=bar\nham=spam\n"



import pytest
from psycopg.errors import UniqueViolation, ForeignKeyViolation


def test_simple_copy(transaction, person):
    if transaction.flavor != "postgresql":
        pytest.skip("COPY FROM not available for sqlite")

    # First copy
    records = [(1, "Big Bob", None), (2, "Bob", None)]
    person.copy_from(records)

    # Check table content
    assert len(list(person.select())) == 2

    # Second copy - must fail on unique index
    records = [(1, "Big Bob", None), (2, "Bob", None)]
    with pytest.raises(UniqueViolation):
        person.copy_from(records)


def test_incorrect_fk(transaction, person):
    if transaction.flavor != "postgresql":
        pytest.skip("COPY FROM not available for sqlite")
    # copy - must fail on fk
    records = [(1, "Trudy", 42)]
    with pytest.raises(ForeignKeyViolation):
        person.copy_from(records)



import pytest

from nagra.transaction import dummy_transaction, Transaction
from nagra.exceptions import NoActiveTransaction
from nagra.schema import Schema


def test_dummy_transaction():

    with pytest.raises(NoActiveTransaction):
        dummy_transaction.execute("SELECT 1")

    with pytest.raises(NoActiveTransaction):
        dummy_transaction.executemany("SELECT 1")


def test_concurrent_transaction(person):
    uri = "postgresql:///nagra"

    with Transaction(uri):
        Schema.default.create_tables()
        # Cleanup
        person.delete()

    trn_a = Transaction(uri)
    trn_b = Transaction(uri)

    person.upsert("name", trn=trn_a).execute("Romeo")
    person.upsert("name", trn=trn_b).execute("Sierra")

    trn_a.commit()
    trn_b.rollback()

    with Transaction(uri) as tr:
        records = list(person.select("name"))
        assert records == [("Romeo",)]

        # Cleanup
        for tbl in Schema.default.tables.values():
            if tbl.is_view:
                continue
            tr.execute(f"DROP TABLE {tbl.name} CASCADE")



import dataclasses
from collections import defaultdict
from collections.abc import Iterable
from functools import partial
from itertools import islice

from nagra.exceptions import UnresolvedFK, ValidationError
from nagra.utils import logger
from nagra.transaction import ExecMany

try:
    from pandas import DataFrame
    from polars import LazyFrame
except ImportError:
    DataFrame = None


class WriterMixin:
    """
    Utility class that provide common methods for Update and Upsert
    """

    def __init__(self):
        self.groups, self.resolve_stm = self.prepare()

    def prepare(self):
        """
        Organise columns in groups and prepare statement to
        resolve fk based on columns expressions
        """
        groups = defaultdict(list)
        for col in self.columns:
            if "." in col:
                head, tail = col.split(".", 1)
                groups[head].append(tail)
            else:
                groups[col] = None

        resolve_stm = {}
        for col, to_select in groups.items():
            if not to_select:
                continue
            cond = ["(= %s {})" % c for c in to_select]
            ftable = self.table.schema.get(self.table.foreign_keys[col])
            select = ftable.select(ftable.primary_key).where(*cond)
            resolve_stm[col] = select.stm()
        return groups, resolve_stm

    def execute(self, *values):
        ids = self.executemany([values])
        if ids:
            return ids[0]

    def executemany(self, records: Iterable[tuple]) -> list:
        # Transform list of records into a dataframe-like dict
        value_df = dict(zip(self.columns, zip(*records)))
        if not value_df:
            return []
        arg_df = {}
        for col, to_select in self.groups.items():
            if to_select:
                values = list(zip(*(value_df[f"{col}.{s}"] for s in to_select)))
                # Try to instanciate lru cache
                cache_key = (self.resolve_stm[col], str(self.lenient))
                lru = self.trn.get_fk_cache(cache_key, fn=partial(self._resolve, col))
                if lru is not None:
                    arg_df[col] = lru.run(values)
                else:
                    arg_df[col] = self._resolve(col, values)
            else:
                arg_df[col] = value_df[col]

        # Build arg iterable
        args = self._exec_args(arg_df)
        # Work by chunks
        stm = self.stm()
        ids = []
        while True:
            chunk = list(islice(args, 1000))
            if not chunk:
                break
            if self.trn.flavor == "sqlite":
                for item in chunk:
                    cursor = self.trn.execute(stm, item)
                    new_id = cursor.fetchone()
                    ids.append(new_id[0] if new_id else None)
            else:
                returning = self.table.primary_key is not None
                cursor = self.trn.executemany(stm, chunk, returning)
                if returning:
                    while True:
                        new_id = cursor.fetchone()
                        ids.append(new_id[0] if new_id else None)
                        if not cursor.nextset():
                            break

        # If conditions are present, enforce those
        if self._check:
            self.validate(ids)
        return ids

    def validate(self, ids: list[int]):
        iter_ids = iter(ids)
        pk = self.table.primary_key
        while True:
            chunk = list(islice(iter_ids, 1000))
            if not chunk:
                return
            cond = self._check + [f"(in {pk} %s)" % (" {}" * len(chunk))]
            select = self.table.select("(count *)").where(*cond)
            (count,) = select.execute(*chunk).fetchone()
            if count != len(chunk):
                msg = f"Validation failed! Condition is: {self._check} )"
                raise ValidationError(msg)

    def _resolve(self, col, values):
        # XXX Detect situation where more than on result is found for
        # a given value (we could also enforce that we only resolve
        # columns with unique constraints) ?
        stm = self.resolve_stm[col]
        exm = ExecMany(stm, values, trn=self.trn)
        for res, vals in zip(exm, values):
            if res is not None:
                yield res[0]
            elif any(v is None for v in vals):
                # One of the values is not given
                yield None
            elif self.lenient is True or col in self.lenient:
                msg = "Value '%s' not found for foreign key column '%s' of table %s"
                logger.info(msg, vals, col, self.table)
                yield None
            else:
                raise UnresolvedFK(
                    f"Unable to resolve '{vals}' (for foreign key "
                    f"{col} of table {self.table.name})"
                )

    def __call__(self, records):
        return self.executemany(records)

    def from_pandas(self, df: "DataFrame"):
        # Convert non-basic types to string
        is_copy = False
        for col in self.columns:
            if df[col].dtype in ("int", "float", "bool", "str"):
                continue
            if not is_copy:
                df = df.copy()
                is_copy = True
            df[col] = df[col].astype(str)

        rows = df[self.columns].values
        return self.executemany(rows)

    def from_polars(self, df: "LazyFrame"):
        from polars import Struct, col

        # Ignore extra columns
        df.select(self.columns)
        schema = df.collect_schema()
        # Convert non-basic types to string
        for name, dtype in schema.items():
            if dtype == Struct:
                df = df.with_columns(col("json").struct.json_encode())

        res = []
        for start, stop in _slicer():
            chunk = df.slice(start, stop).collect()
            if chunk.is_empty():
                break
            rows = chunk.iter_rows()
            res += self.executemany(rows)
        return res

    def from_dict(self, records):
        # Create select object in order to generate the same column names
        select = self.table.select(*self.columns)
        field_names = [f.name for f in dataclasses.fields(select.to_dataclass())]

        # Extract dict values - allows for field names or dotted column format
        f_or_c = zip(field_names, self.columns)
        rows = (tuple(getter(record, field, col) for col, field in f_or_c) for record in records)
        return self.executemany(rows)


def getter(record, field, col):
    """
    Utility function to extract value as a column name (with
    potentially dots) or a field name from record
    """
    if field in record:
        return record[field]
    if col in record:
        return record[col]
    raise KeyError(f"KeyError: neither {field} or {col} found")

def _slicer(chunk_size=10_000):
    start = 0
    while True:
        stop = start + chunk_size
        yield start, stop
        start = stop



import re
from collections.abc import Iterable
from dataclasses import dataclass, make_dataclass, fields as dataclass_fields
from datetime import datetime, date
from itertools import islice, repeat, takewhile
from typing import Optional, Union, TYPE_CHECKING

from nagra import Statement, Schema
from nagra.exceptions import ValidationError
from nagra.sexpr import AST, AggToken
from nagra.utils import snake_to_pascal, get_table_from_dataclass, iter_dataclass_cols

if TYPE_CHECKING:
    from nagra.table import Env, Table
    from nagra.transaction import Transaction
    from pandas import DataFrame
    from polars import LazyFrame

RE_VALID_IDENTIFIER = re.compile(r"\W|^(?=\d)")


def clean_col(name):
    return RE_VALID_IDENTIFIER.sub("_", name)


class Select:
    def __init__(self, table: "Table", *columns: str, trn: "Transaction", env: "Env"):
        self.table = table
        self.env = env
        self.where_asts = tuple()
        self._offset = None
        self._limit = None
        self._aliases = tuple()
        self.groupby_ast = tuple()
        self.order_ast = tuple()
        self.order_directions = tuple()
        self.columns = tuple()
        self.columns_ast = tuple()
        self.query_columns = tuple()
        self.trn = trn
        self._add_columns(columns)

    def _add_columns(self, columns):
        self.columns += columns
        self.columns_ast += tuple(AST.parse(c) for c in columns)
        self.query_columns += tuple(
            a.eval(self.env, self.trn.flavor) for a in self.columns_ast
        )

    def clone(self, trn: Optional["Transaction"] = None):
        """
        Return a copy of select with updated parameters
        """
        trn = trn or self.trn
        cln = Select(self.table, *self.columns, trn=trn, env=self.env.clone())
        cln.where_asts = self.where_asts
        cln.groupby_ast = self.groupby_ast
        cln.order_ast = self.order_ast
        cln.order_directions = self.order_directions
        cln._limit = self._limit
        cln._offset = self._offset
        cln._aliases = self._aliases
        return cln

    def where(self, *conditions: str):
        cln = self.clone()
        cln.where_asts += tuple(AST.parse(cond) for cond in conditions)
        return cln

    def aliases(self, *names: str):
        cln = self.clone()
        cln._aliases += names
        return cln

    def select(self, *columns: str):
        cln = self.clone()
        cln._add_columns(columns)
        return cln

    def offset(self, value):
        cln = self.clone()
        cln._offset = value
        return cln

    def limit(self, value):
        cln = self.clone()
        cln._limit = value
        return cln

    def groupby(self, *groups: str):
        cln = self.clone()
        cln.groupby_ast += tuple(AST.parse(g) for g in groups)
        return cln

    def orderby(self, *orders: str | tuple[str, str]):
        expressions = []
        directions = []
        for o in orders:
            if isinstance(o, tuple):
                expression = o[0]
                direction = o[1]
            else:
                expression = o
                direction = "asc"

            if isinstance(expression, int):
                expression = self.columns[expression]
            expressions.append(expression)
            directions.append(direction)

        cln = self.clone()
        cln.order_ast += tuple(AST.parse(e) for e in expressions)
        cln.order_directions += tuple(directions)
        return cln

    def to_dataclass(self, *aliases: str, model_name=None, nest=False) -> dataclass:
        fields = {}
        model_name = model_name or snake_to_pascal(self.table.name)
        for col, (name, dt) in zip(self.columns, self.dtypes(*aliases)):
            # Handle nested models
            if nest and "." in col:
                head, _ = col.split(".", 1)
                if head in fields:
                    # Fields for this fk already managed
                    continue
                # Create nested fields
                if head in self.table.foreign_keys:
                    fname = self.table.foreign_keys[head]
                else:
                    alias = self.table.one2many[head]
                    fname, _ = alias.split(".")
                ftable = self.table.schema.get(fname)
                tails = []
                for sub_col in self.columns:
                    prefix = f"{head}."
                    if sub_col.startswith(prefix):
                        tails.append(sub_col.removeprefix(prefix))
                # trigger a select on foreign table and generate dataclass
                sub_class = ftable.select(*tails).to_dataclass(
                    nest=True,
                    model_name=model_name + snake_to_pascal(head)
                )
                nullable = not self.table.required(head)
                if nullable:
                    sub_class = sub_class | None
                field_def = (
                    head,
                    sub_class,
                )
                if nullable:
                    field_def += (None,)
                fields[head] = field_def

            # Handle required/optional fields
            else:
                field_def = (
                    clean_col(name),
                    dt,
                )
                if not self.table.required(name):
                    field_def += (None,)
                fields[name] = field_def

        return make_dataclass(
            model_name, fields=fields.values(), kw_only=True
        )

    @staticmethod
    def from_dataclass(cls: dataclass, schema: Schema = Schema.default) -> "Select":
        # TODO add from_pydantic
        table = get_table_from_dataclass(cls, schema)
        cols = list(iter_dataclass_cols(cls))
        return table.select(*cols)

    def to_pydantic(self, *aliases: str, model_name=None):
        from pydantic import create_model

        fields = {}
        for name, dt in self.dtypes(*aliases):
            cleaned = clean_col(name)
            if self.table.required(name):
                field_def = (dt, ...)
            else:
                # Add default value
                field_def = (dt, None)
            fields[cleaned] = field_def

        return create_model(model_name or self.table.name, **fields)

    def dtypes(self, *aliases: str, with_optional: bool = True):
        fields = []
        if aliases:
            assert len(aliases) == len(self.columns)
        else:
            aliases = self.columns

        # Construct fields
        for alias, col_name, col_ast in zip(aliases, self.columns, self.columns_ast):
            # Eval type and nullable
            col_type = col_ast.eval_type(self.env)
            nullable = col_ast.is_nullable(self.env)
            # Eval nullable
            if with_optional and nullable:
                # Fixme Optional may depend on ast content
                col_type = col_type | None
            fields.append((alias, col_type))
        return fields

    def infer_groupby(self):
        # Detect aggregates
        for a in self.columns_ast:
            if any(isinstance(tk, AggToken) for tk in a.chain()):
                break
        else:
            # No aggregate found
            return []

        # Collect non-aggregates
        groupby_ast = []
        for a in self.columns_ast:
            if any(isinstance(tk, AggToken) for tk in a.chain()):
                continue
            groupby_ast.append(a)
        return groupby_ast

    def stm(self):
        # Eval where conditions
        where_conditions = [
            ast.eval(self.env, self.trn.flavor) for ast in self.where_asts
        ]
        # Eval Groupby
        groupby_ast = self.groupby_ast or self.infer_groupby()
        groupby = [a.eval(self.env, self.trn.flavor) for a in groupby_ast]
        # Eval Oder by
        orderby = [
            a.eval(self.env, self.trn.flavor) + f" {d}"
            for a, d in zip(
                self.order_ast,
                self.order_directions,
            )
        ]
        # Create joins
        joins = self.table.join(self.env)

        if self._aliases:
            query_columns = [
                f"{c} AS {a}" for c, a in zip(self.query_columns, self._aliases)
            ]
        else:
            query_columns = self.query_columns

        stm = Statement(
            "select",
            table=self.table.name,
            columns=query_columns,
            joins=joins,
            conditions=where_conditions,
            limit=self._limit,
            offset=self._offset,
            groupby=groupby,
            orderby=orderby,
        )
        return stm()

    def to_polars(self, *args) -> "LazyFrame":
        assert self.trn.flavor != "sqlite", "Polars is only supported with Postgresql"
        import polars

        schema = self.dtypes(with_optional=False)
        cursor = self.execute(*args)
        columns = [c for c, _ in schema]
        df = polars.LazyFrame(cursor, schema=columns)
        if self._aliases:
            mapping = dict(zip((n for n, _ in schema), self._aliases))
            df = df.rename(mapping)
        return df

    def to_pandas(
        self, *args, chunked: int = 0
    ) -> Union["DataFrame", Iterable["DataFrame"]]:
        """
        Execute the query with given args and return a pandas
        DataFrame. If chunked is bigger than 0, return an iterable
        yielding dataframes.
        """
        names, dtypes = zip(*(self.dtypes(with_optional=False)))
        cursor = self.execute(*args)
        if chunked <= 0:
            return self.create_df(cursor, names, dtypes)

        # Create generator
        chunkify = (list(islice(cursor, chunked)) for _ in repeat(None))
        # Return df as long as the generator yield non-empty list
        return (
            self.create_df(chunk, names, dtypes) for chunk in takewhile(bool, chunkify)
        )

    def create_df(self, cursor: Iterable[tuple], names: tuple[str, ...], dtypes: tuple):
        """
        Create a Dataframe, whose columns name are `names` and
        types `dtypes`.
        """
        from pandas import DataFrame, Series, to_datetime

        df = DataFrame()
        by_col = zip(*cursor)
        for name, dt, col in zip(names, dtypes, by_col):
            # FIXME Series(col, dtype=dt) fail on json cols!
            srs = Series(col)
            if dt in (datetime, date):
                srs = to_datetime(srs)
            else:
                if dt == int:
                    # Make sure we have no nan for int columns
                    srs = srs.fillna(0)
                try:
                    srs = srs.astype(dt)
                except TypeError:
                    # Fallback to string if type is not supported by pandas
                    srs = srs.astype(str)
            df[name] = srs

        if df.columns.empty:
            # No records were returned by the cursor
            df = DataFrame(columns=names)
        if self._aliases:
            df.columns = self._aliases
        return df

    def to_dict(self, *args, nest=False) -> Iterable[dict]:
        if nest:
            if self._aliases:
                msg = "Nesting and fields aliases can not be combined"
                raise ValidationError(msg)
            yield from self.to_nested_dict(*args)
        else:
            columns = [f.name for f in dataclass_fields(
                self.to_dataclass(*self._aliases)
            )]
            for row in self.execute(*args):
                yield dict(zip(columns, row))

    def to_nested_dict(self, *args) -> Iterable[dict]:
        for row in self.execute(*args):
            record = dict(zip(self.columns, row))
            yield autonest(record)

    def execute(self, *args):
        return self.trn.execute(self.stm(), args)

    def executemany(self, args):
        return self.trn.executemany(self.stm(), args)

    def one(self, *args):
        return self.trn.execute(self.stm(), args).fetchone()

    def __iter__(self):
        return iter(self.execute())


def autonest(record: dict) -> dict:
    clone = {}
    for key, value in record.items():
        if "." not in key:
            clone[key] = value
            continue
        head, _ = key.split(".", 1)
        prefix = f"{head}."
        sub_dict = {
            k.removeprefix(prefix): v
            for k, v in record.items()
            if k.startswith(prefix)
        }
        if all(v is None for v in sub_dict.values()):
            # We only collect sub_dict if not fully null
            clone[head] = None
            continue
        clone[head] = autonest(sub_dict)
    return clone



from nagra import Statement
from nagra.utils import strip_lines
from nagra.schema import Schema
from nagra.table import Table


def test_create_table(empty_transaction):
    flavor = empty_transaction.flavor
    schema = Schema()
    Table(
        "my_table",
        columns={
            "name": "varchar",
            "score": "int",
        },
        natural_key=["name"],
        not_null=["score"],
        default={
            "score": "0",
        },
        primary_key="custom_id",
        schema=schema,
    )
    lines = list(schema.setup_statements(trn=empty_transaction))
    create_table, add_name, add_score, create_idx = map(strip_lines, lines)

    if flavor == "postgresql":
        assert create_table == [
            'CREATE TABLE  "my_table" (',
            '"custom_id" BIGSERIAL PRIMARY KEY',
            ");",
        ]
    else:
        assert create_table == [
            'CREATE TABLE  "my_table" (',
            '"custom_id"  INTEGER PRIMARY KEY',
            ");",
        ]
    assert add_name == ['ALTER TABLE "my_table"', 'ADD COLUMN "name" TEXT NOT NULL']

    assert add_score == [
        'ALTER TABLE "my_table"',
        'ADD COLUMN "score" INTEGER NOT NULL',
        "DEFAULT 0",
    ]
    assert create_idx == [
        'CREATE UNIQUE INDEX my_table_idx ON "my_table" (',
        '"name"',
        ");",
    ]


def test_create_table_pk_is_fk(empty_transaction):
    flavor = empty_transaction.flavor
    schema = Schema()
    Table(  # Regular table
        "concept",
        columns={
            "name": "varchar",
        },
        natural_key=["name"],
        primary_key="concept_id",
        schema=schema,
    )
    Table(  # Table with no primary key and a fk in the nk
        "score",
        columns={
            "concept": "bigint",
            "score": "int",
        },
        primary_key="concept",
        foreign_keys={
            "concept": "concept",
        },
        schema=schema,
    )
    lines = list(schema.setup_statements(trn=empty_transaction))
    if flavor == "postgresql":
        assert lines == [
            'CREATE TABLE  "concept" (\n  "concept_id" BIGSERIAL PRIMARY KEY\n);',
            'CREATE TABLE  "score" (\n'
            '  "concept" BIGINT PRIMARY KEY\n'
            '   CONSTRAINT fk_concept REFERENCES "concept"("concept_id")\n'
            ");",
            'ALTER TABLE "concept"\n ADD COLUMN "name" TEXT NOT NULL',
            'ALTER TABLE "score"\n ADD COLUMN "score" INTEGER NOT NULL',
            'CREATE UNIQUE INDEX concept_idx ON "concept" (\n  "name"\n);',
            'CREATE UNIQUE INDEX score_idx ON "score" (\n  "concept", "score"\n);',
        ]
    else:
        assert lines == [
            'CREATE TABLE  "concept" (\n  "concept_id"  INTEGER PRIMARY KEY\n);',
            'CREATE TABLE  "score" (\n'
            '  "concept"  INTEGER PRIMARY KEY\n'
            '   CONSTRAINT fk_concept REFERENCES "concept"("concept_id")\n'
            ");",
            'ALTER TABLE "concept"\n ADD COLUMN "name" TEXT NOT NULL\n',
            'ALTER TABLE "score"\n ADD COLUMN "score" INTEGER NOT NULL\n',
            'CREATE UNIQUE INDEX concept_idx ON "concept" (\n  "name"\n);',
            'CREATE UNIQUE INDEX score_idx ON "score" (\n  "concept", "score"\n);',
        ]


def test_create_table_no_pk(empty_transaction):
    flavor = empty_transaction.flavor
    schema = Schema()
    Table(  # Regular table
        "concept",
        columns={
            "name": "varchar",
        },
        natural_key=["name"],
        schema=schema,
    )
    Table(  # Table with no primary key and a fk in the nk
        "score",
        columns={
            "concept": "bigint",
            "score": "int",
        },
        natural_key=["concept"],
        primary_key=None,
        foreign_keys={
            "concept": "concept",
        },
        schema=schema,
    )
    lines = list(schema.setup_statements(trn=empty_transaction))
    (
        create_concept,
        create_score_table,
        add_concept_name,
        add_score,
        create_concept_idx,
        create_score_idx,
    ) = map(strip_lines, lines)

    if flavor == "postgresql":
        assert create_concept == [
            'CREATE TABLE  "concept" (',
            '"id" BIGSERIAL PRIMARY KEY',
            ");",
        ]

        assert create_score_table == [
            'CREATE TABLE  "score" (',
            '"concept"  BIGINT NOT NULL',
            'CONSTRAINT fk_concept REFERENCES "concept"("id")',
            ");",
        ]
    else:
        assert create_concept == [
            'CREATE TABLE  "concept" (',
            '"id"  INTEGER PRIMARY KEY',
            ");",
        ]

        assert create_score_table == [
            'CREATE TABLE  "score" (',
            '"concept"  INTEGER NOT NULL',
            'CONSTRAINT fk_concept REFERENCES "concept"("id")',
            ");",
        ]
    assert add_concept_name == [
        'ALTER TABLE "concept"',
        'ADD COLUMN "name" TEXT NOT NULL',
    ]
    assert add_score == ['ALTER TABLE "score"', 'ADD COLUMN "score" INTEGER']
    assert create_concept_idx == [
        'CREATE UNIQUE INDEX concept_idx ON "concept" (',
        '"name"',
        ");",
    ]
    assert create_score_idx == [
        'CREATE UNIQUE INDEX score_idx ON "score" (',
        '"concept"',
        ");",
    ]


def test_create_unique_index():
    stmt = Statement("create_unique_index").table("my_table").natural_key(["name"])
    doc = stmt()
    lines = strip_lines(doc)
    assert lines == [
        'CREATE UNIQUE INDEX my_table_idx ON "my_table" (',
        '"name"',
        ");",
    ]



import zoneinfo
from datetime import datetime, date
from uuid import UUID

import polars
import pytest


def test_to_polars(transaction, temperature):
    if transaction.flavor != "postgresql":
        pytest.skip("Sqlite not supported with polars")

    # Upsert
    temperature.upsert("timestamp", "city", "value").executemany(
        [
            ("1970-01-02", "Berlin", 10),
            ("1970-01-02", "London", 12),
        ]
    )
    # Read data
    df = temperature.select().to_polars()
    columns = df.collect_schema().names()
    assert columns == ["timestamp", "city", "value"]
    # to_polars return a LazyFrame, we matertialize it
    df = df.collect()
    assert sorted(df['city']) == ["Berlin", "London"]

    # Read with custom arg
    cond = "(= value {})"
    df = temperature.select().where(cond).to_polars(12).collect()
    assert list(df.columns) == ["timestamp", "city", "value"]
    assert sorted(df['city']) == ["London"]


def test_from_polars(transaction, kitchensink):
    if transaction.flavor != "postgresql":
        pytest.skip("Sqlite not supported with polars")
    df = polars.LazyFrame(
        {
            "varchar": ["ham"],
            "bigint": [1],
            "float": [1.0],
            "int": [1],
            "timestamp": ["1970-01-01 00:00:00"],
            "timestamptz": ["1970-01-01 00:00:00+00:00"],
            "bool": [True],
            "date": ["1970-01-01"],
            "json": [{}],
            "uuid": ["F1172BD3-0A1D-422E-8ED6-8DC2D0F8C11C"],
            "max": ["max"],
            "true": ["true"],
            "blob": [b"blob"],
        }
    )

    # UPSERT
    kitchensink.upsert().from_polars(df)
    (row,) = kitchensink.select()
    BRUTZ = zoneinfo.ZoneInfo(key="Europe/Brussels")
    assert row == (
        "ham",
        1,
        1.0,
        1,
        datetime(1970, 1, 1, 0, 0),
        datetime(1970, 1, 1, 1, 0, tzinfo=BRUTZ),
        True,
        date(1970, 1, 1),
        {},
        UUID("F1172BD3-0A1D-422E-8ED6-8DC2D0F8C11C"),
        "max",
        "true",
        b"blob",
    )

    # SELECT with operator
    new_df = kitchensink.select(
        "(date_bin '5 days' timestamptz '1900-01-01')",
    ).to_polars().collect()

    new_df.columns = ["ts"]
    ts = new_df['ts']
    assert str(ts.dtype) == "Datetime(time_unit='us', time_zone='Europe/Brussels')"

    assert ts[0].isoformat() == '1969-12-30T01:00:00+01:00'



from collections.abc import Iterable
from typing import Optional, TYPE_CHECKING

from nagra import Statement
from nagra.sexpr import AST

if TYPE_CHECKING:
    from nagra.table import Table, Env
    from nagra.transaction import Transaction


class Delete:
    def __init__(
        self, table: "Table", trn: "Transaction", env: "Env", where: Iterable[str] = []
    ):
        self.table = table
        self.trn = trn
        self.env = env
        self._where = list(where)

    def clone(
        self,
        trn: Optional["Transaction"] = None,
        env: Optional["Env"] = None,
        where: Iterable[str] = [],
    ):
        """
        Return a copy of upsert with updated parameters
        """
        trn = trn or self.trn
        where = self._where + list(where)
        cln = Delete(
            self.table,
            env=env or self.env,
            trn=trn,
            where=where,
        )
        return cln

    def where(self, *conditions: str):
        return self.clone(where=conditions)

    def stm(self):
        asts = [AST.parse(cond) for cond in self._where]
        eval_conditions = [ast.eval(self.env, flavor=self.trn.flavor) for ast in asts]
        joins = list(self.table.join(self.env))
        stm = Statement(
            "delete-with-join" if joins else "delete",
            table=self.table.name,
            joins=joins,
            conditions=eval_conditions,
        )
        return stm()

    def __call__(self):
        return self.execute()

    def execute(self, *args):
        return self.trn.execute(self.stm(), args)

    def executemany(self, args):
        return self.trn.executemany(self.stm(), args)

    def __iter__(self):
        return iter(self.execute())



"""
The `AST` (abstract syntax tree) implement parsing and evaluation
of [s-expressions](https://en.wikipedia.org/wiki/S-expression).

Example:

``` python-console
>>> from nagra.sexpr import AST
>>> AST.parse('(+ 1 1)')
<nagra.sexpr.AST object at 0x7f1bcc5b2fd0>
>>> ast = AST.parse('(+ 1 1)')
>>> ast.eval()
2
```

The `eval` method accepts an `env` parameter, a dictionary used to
evaluate non-litteral tokens:

``` python-console
>>> ast = AST.parse('(+ 1 x)')
>>> ast.eval()
Traceback (most recent call last):
   ...
ValueError: Unexpected token: "x"
>>> ast.eval({'x': 2})
3
```

"""

import shlex
from datetime import date, datetime
from functools import cached_property

from nagra.exceptions import EvalTypeError


DEFAULT_FLAVOR = "postgresql"
__all__ = ["AST"]


def list_to_dict(*items):
    it = iter(items)
    return dict(zip(it, it))


class KWargs:
    def __init__(self, *items):
        self.value = list_to_dict(*items)

    def __repr__(self):
        return f'<KWargs "{self.value}">'


class Alias:
    """
    Simple wrapper that combine a value and an alias name
    """

    def __init__(self, value, name):
        self.value = value
        self.name = name


def tokenize(expr):
    lexer = shlex.shlex(expr)
    lexer.wordchars += ".!=<>:{}-|"
    prev_tk = None
    for i in lexer:
        tk = Token.from_value(i, prev_tk)
        prev_tk = tk
        yield tk


def scan(tokens, end_tk=")"):
    res = []
    for tk in tokens:
        if tk.value == end_tk:
            return res
        elif tk.value == "(":
            res.append(scan(tokens))
        else:
            res.append(tk)

    tail = next(tokens, None)
    if tail:
        raise ValueError(f'Unexpected token: "{tail.value}"')
    return res


class AST:
    # Litterals
    literals = {
        "true",
        "false",
        "null",
        "*",
    }

    # Set of infix operators
    infix = {
        "and",
        "not",
        "is",
        "in",
        "+",
        "-",
        "*",
        "/",
        "||",
        "=",
        "!=",
        "<",
        "<=",
        ">",
        ">=",
        "is",
        "match",
        "ilike",
        "like",
        "||",
        "&",
        "|",
        "#",
        "~",
        "<<",
        ">>",
        "->",
        "->>",
        "#>",
        "#>>",
        "<->",
        "<#>",
        "<=>",
        "<+>",
        "<~>",
        "<%>",
    }
    # Declare special builtins  explicitly
    builtins = {
        "-": (lambda *xs: " - ".join(map(str, xs)) if len(xs) > 1 else f"-{xs[0]}"),
        "not": "NOT {}".format,
        # OR is explicitly declared because of operator precedence
        "or": lambda *x: "(%s)" % (" OR ".join(x)),
        "isnot": "NOT {} IS {}".format,
        "extract": "EXTRACT({} FROM {})".format,
        "in": lambda x, *ys: f"{x} in (%s)" % ", ".join(map(str, ys)),
        # Usefull for IN operation with large number of items (see
        # https://postgres.cz/wiki/PostgreSQL_SQL_Tricks_I#Predicate_IN_optimalization)
        "values": lambda *xs: ("(VALUES %s)" % (
            ",".join("({})" for _ in xs)
        )).format(*xs)
    }

    # Declare aggregate operators
    agg_unary = {
        "min",
        "max",
        "sum",
        "avg",
        "every",
        "count",
        "group_concat",
        "string_agg",
        "array_agg",
        "json_agg",
        "bool_or",
        "bool_and",
        "json_object_agg",
    }
    agg_variadic = {
        "group_concat",
        "string_agg",
    }
    aggregates = agg_unary | agg_variadic

    def __init__(self, tokens):
        # Auto-wrap sublist into AST
        self.tokens = [tk if isinstance(tk, Token) else AST(tk) for tk in tokens]

    @classmethod
    def parse(cls, expr):
        res = tokenize(expr)
        tokens = scan(res)[0]
        if isinstance(tokens, Token):
            tokens = [tokens]
        return AST(tokens)

    def chain(self):
        for tk in self.tokens:
            if isinstance(tk, Token):
                yield tk
            else:
                yield from tk.chain()

    def _eval(self, env, flavor, top=False):
        head, tail = self.tokens[0], self.tokens[1:]
        args = [tk._eval(env, flavor) for tk in tail]
        res = head._eval(env, flavor, *args)
        return res if top else "({})".format(res)

    def eval(self, env, flavor=DEFAULT_FLAVOR):
        return self._eval(env, flavor, top=True)

    def relations(self):
        for tk in self.chain():
            if tk.is_relation():
                yield tk.value

    def _eval_type(self, env):
        # head is always a token
        head, tail = self.tokens[0], self.tokens[1:]
        args = [tk._eval_type(env) for tk in tail]
        res = head._eval_type(env, *args)
        return res

    def eval_type(self, env):
        return self._eval_type(env)

    def is_nullable(self, env):
        return any(
            tk._is_nullable(env)
            for tk in self.chain()
            if not isinstance(tk, (BuiltinToken, ))
        )

    def get_args(self):
        """
        Return token that should be treated as query arguments
        """
        args = list((tk.get_arg() for tk in self.chain()))
        return [a for a in args if a]

    def is_aggregate(self):
        for tk in self.chain():
            if tk.is_aggregate():
                return True
        return False


class Token:
    def __init__(self, value):
        self.value = value

    def is_relation(self):
        return False

    @staticmethod
    def from_value(value, prev_tk):
        # Handle exceptions
        if value == "(":
            return LParen(value)

        # Dot prefix force Vartoken
        if value.startswith("."):
            return VarToken(value.lstrip("."))

        # Literals take precedence over everything else
        if value in AST.literals:
            return LiteralToken(value)

        # First item should be an operator
        if isinstance(prev_tk, LParen):
            if value in AST.literals:
                return LiteralToken(value)
            elif value in AST.aggregates:
                return AggToken(value)
            else:
                return BuiltinToken(value)

        if (value[0], value[-1]) == ("{", "}"):
            return ParamToken(value)
        try:
            if "." in value:
                value = float(value)
                return FloatToken(value)
            else:
                value = int(value)
                return IntToken(value)
        except ValueError:
            pass
        return StrToken(value) if value[0] in "\"'" else VarToken(value)

    def __repr__(self):
        cls = self.__class__.__name__
        return f"<{cls} {self.value}>"

    def get_arg(self):
        return None

    def _eval(self, env, flavor, *args):
        return None


class LParen(Token):
    "Left Parenthesis"


class ParamToken(Token):
    "Parameterized Token"

    def __init__(self, value):
        # Remove braces
        self.value = value[1:-1]
        # TODO self.value placeholder name we should use it to apply
        # param, for example when we do:
        # select.where('(= col {my_input})').execute(my_input=42)

    def _eval(self, env, flavor, *args):
        placeholder = "%s" if flavor == "postgresql" else "?"
        return placeholder


class VarToken(Token):
    def __init__(self, value):
        self.join_alias = None
        super().__init__(value)

    def is_relation(self):
        return "." in self.value

    def _eval(self, env, flavor, *args):
        if self.is_relation():
            self.join_alias = env.add_ref(self.value.split("."))
            return self.join_alias
        return '"{}"."{}"'.format(env.table.name, self.value)

    def _eval_type(self, env):
        # TODO handle paramtoken here?
        if self.is_relation():
            *head, tail = self.value.split(".")
            ftable, _, _ = env.table.join_on(tuple(head), env=env)
            if tail not in ftable.columns and tail == ftable.primary_key:
                # implicit type for pk is int
                # FIXME this would be simpler with id in columns
                return int
            col = ftable.columns[tail]
            return col.python_type()
        elif col := env.table.columns.get(self.value):
            return col.python_type()
        elif self.value == env.table.primary_key:
            # implicit type for pk is integer
            return int
        else:
            raise EvalTypeError(f"Unable to eval type of '{self.value}'")

    def _is_nullable(self, env):
        if self.is_relation():
            value = self.value
            table = env.table
            while "." in value:
                # If any item in the chain is nullable, the all chain is
                head, value = value.split(".", 1)
                if head not in table.not_null:
                    return True
                table = table.schema.get(table.foreign_keys[head])
            # No nullable column in the dotted chain
            return False
        return self.value not in env.table.not_null


class OpToken(Token):
    @cached_property
    def op(self):
        if self.value in AST.builtins:
            return AST.builtins[self.value]
        if self.value in AST.infix:
            return lambda *xs: f" {self.value.upper()} ".join(map(str, xs))
        return lambda *xs: f"{self.value}(%s)" % ", ".join(map(str, xs))

    def _eval(self, env, flavor, *args):
        if self.value in AST.literals:
            return self.value
        return self.op(*args)


class BuiltinToken(OpToken):
    num_like = set(['+', '-', '*', '/'])
    bool_like = set([
        "!=",
        "<",
        ">",
        ">=",
        "<=",
        "=",
        "and",
        "or",
        "not",
        "is",
        "isnot",
        "like",
        "ilike",
        "isfinite",
    ])
    datetime_like = set([
        "clock_timestamp",
        "current_timestamp",
        "date_add",
        "date_bin",
        "date_subtract",
        "date_trunc",
        "localtimestamp",
        "make_timestamp",
        "now",
        "statement_timestamp",
        "to_timestamp",
        "transaction_timestamp",
    ])
    date_like = set([
        "current_date",
        "make_date",
    ])
    float_like = set([
        "date_part",
        "extract",
    ])

    def _eval_type(self, env, *operands):
        # FIXME, probably too basic
        if self.value in self.num_like:
            if all(op == date for op in operands):
                return int
            elif any(op == date for op in operands):
                return date
            elif any(op == float for op in operands):
                return float
            return int
        elif self.value in self.bool_like:
            return bool
        elif self.value in self.float_like:
            return float
        elif self.value in self.datetime_like:
            return datetime
        elif self.value in self.date_like:
            return date
        else:
            return str


class LitToken(Token):
    "Litteral Token"

    def _eval_type(self, env):
        return self._type

    def _eval(self, env, flavor, *args):
        return self.value

    def _is_nullable(self, env):
        return False


class FloatToken(LitToken):
    "Float Token"
    _type = float


class IntToken(LitToken):
    "Integer Token"
    _type = int


class StrToken(LitToken):
    "String Token"

    def __init__(self, value):
        # Remove quotes
        self.value = value[1:-1]

    def _eval_type(self, env):
        return str

    def _eval(self, env, flavor, *args):
        return f"'{self.value}'"


class LiteralToken(OpToken):
    """
    Class for hard-coded litteral token, one of `AST.literals`
    """
    def _eval_type(self, env, *operands):
        if self.value == "null":
            return None
        return bool

    def _is_nullable(self, env, *operands):
        return self.value == "null"


class AggToken(OpToken):
    ops = AST.aggregates

    num_like = ["sum", "avg"]
    bool_like = ["every"]

    def _eval_type(self, env, *operands):
        if self.value == "count":
            return int
        if self.value in self.num_like:
            assert operands[0] in (float, int, list[float], list[int])
            return operands[0]
        if self.value in self.bool_like:
            assert operands[0] == bool
            return bool
        else:
            return operands[0]

    def _is_nullable(self, env, *operands):
        return True
